/*
 * Souffle - A Datalog Compiler
 * Copyright (c) 2017, The Souffle Developers. All rights reserved.
 * Licensed under the Universal Permissive License v 1.0 as shown at:
 * - https://opensource.org/licenses/UPL
 * - <souffle root>/licenses/SOUFFLE-UPL.txt
 */

/************************************************************************
 *
 * @file ProvenanceTransformer.cpp
 *
 * Implements AstTransformer for adding provenance information via two extra columns
 *
 ***********************************************************************/

#include "AstTransforms.h"

namespace souffle {

/**
 * Helper functions
 */
const std::string identifierToString(const AstRelationIdentifier& name) {
    std::stringstream ss;
    ss << name;
    return ss.str();
}

inline AstRelationIdentifier makeRelationName(
        const AstRelationIdentifier& orig, const std::string& type, int num = -1) {
    AstRelationIdentifier newName(identifierToString(orig));
    newName.append(type);
    if (num != -1) {
        newName.append((const std::string&)std::to_string(num));
    }

    return newName;
}

std::unique_ptr<AstRelation> makeInfoRelation(
        AstClause& originalClause, AstTranslationUnit& translationUnit) {
    AstRelationIdentifier name =
            makeRelationName(originalClause.getHead()->getName(), "@info", originalClause.getClauseNum());

    // initialise info relation
    auto infoRelation = new AstRelation();
    infoRelation->setName(name);

    // create new clause containing a single fact
    auto infoClause = new AstClause();
    auto infoClauseHead = new AstAtom();
    infoClauseHead->setName(name);

    infoRelation->addAttribute(
            std::unique_ptr<AstAttribute>(new AstAttribute("clause_num", AstTypeIdentifier("number"))));
    infoClauseHead->addArgument(
            std::unique_ptr<AstArgument>(new AstNumberConstant(originalClause.getClauseNum())));

    // visit all body literals and add to info clause head
    for (size_t i = 0; i < originalClause.getBodyLiterals().size(); i++) {
        auto lit = originalClause.getBodyLiterals()[i];
        const AstAtom* atom = lit->getAtom();
        if (atom != nullptr) {
            std::string relName = identifierToString(atom->getName());

            infoRelation->addAttribute(std::unique_ptr<AstAttribute>(
                    new AstAttribute(std::string("rel_") + std::to_string(i), AstTypeIdentifier("symbol"))));

            if (dynamic_cast<AstAtom*>(lit)) {
                infoClauseHead->addArgument(std::unique_ptr<AstArgument>(
                        new AstStringConstant(translationUnit.getSymbolTable(), relName.c_str())));
            } else if (dynamic_cast<AstNegation*>(lit)) {
                infoClauseHead->addArgument(std::unique_ptr<AstArgument>(
                        new AstStringConstant(translationUnit.getSymbolTable(), ("!" + relName).c_str())));
            }
        }
    }

    // generate and add clause representation
    std::stringstream ss;
    originalClause.print(ss);

    infoRelation->addAttribute(
            std::unique_ptr<AstAttribute>(new AstAttribute("clause_repr", AstTypeIdentifier("symbol"))));
    infoClauseHead->addArgument(std::unique_ptr<AstArgument>(
            new AstStringConstant(translationUnit.getSymbolTable(), ss.str().c_str())));

    // set clause head and add clause to info relation
    infoClause->setHead(std::unique_ptr<AstAtom>(infoClauseHead));
    infoRelation->addClause(std::unique_ptr<AstClause>(infoClause));

    return std::unique_ptr<AstRelation>(infoRelation);
}

/** Transform eqrel relations to explicitly define equivalence relations */
void transformEqrelRelation(AstRelation& rel) {
    assert(rel.isEqRel() && "attempting to transform non-eqrel relation");
    assert(rel.getArity() == 2 && "eqrel relation not binary");

    rel.setQualifier(rel.getQualifier() - EQREL_RELATION);

    // transitivity
    // transitive clause: A(x, z) :- A(x, y), A(y, z).
    auto transitiveClause = new AstClause();
    auto transitiveClauseHead = new AstAtom(rel.getName());
    transitiveClauseHead->addArgument(std::unique_ptr<AstArgument>(new AstVariable("x")));
    transitiveClauseHead->addArgument(std::unique_ptr<AstArgument>(new AstVariable("z")));

    auto transitiveClauseBody = new AstAtom(rel.getName());
    transitiveClauseBody->addArgument(std::unique_ptr<AstArgument>(new AstVariable("x")));
    transitiveClauseBody->addArgument(std::unique_ptr<AstArgument>(new AstVariable("y")));

    auto transitiveClauseBody2 = new AstAtom(rel.getName());
    transitiveClauseBody2->addArgument(std::unique_ptr<AstArgument>(new AstVariable("y")));
    transitiveClauseBody2->addArgument(std::unique_ptr<AstArgument>(new AstVariable("z")));

    transitiveClause->setHead(std::unique_ptr<AstAtom>(transitiveClauseHead));
    transitiveClause->addToBody(std::unique_ptr<AstLiteral>(transitiveClauseBody));
    transitiveClause->addToBody(std::unique_ptr<AstLiteral>(transitiveClauseBody2));
    rel.addClause(std::unique_ptr<AstClause>(transitiveClause));

    // symmetric
    // symmetric clause: A(x, y) :- A(y, x).
    auto symClause = new AstClause();
    auto symClauseHead = new AstAtom(rel.getName());
    symClauseHead->addArgument(std::unique_ptr<AstArgument>(new AstVariable("x")));
    symClauseHead->addArgument(std::unique_ptr<AstArgument>(new AstVariable("y")));

    auto symClauseBody = new AstAtom(rel.getName());
    symClauseBody->addArgument(std::unique_ptr<AstArgument>(new AstVariable("y")));
    symClauseBody->addArgument(std::unique_ptr<AstArgument>(new AstVariable("x")));

    symClause->setHead(std::unique_ptr<AstAtom>(symClauseHead));
    symClause->addToBody(std::unique_ptr<AstLiteral>(symClauseBody));
    rel.addClause(std::unique_ptr<AstClause>(symClause));

    // reflexivity
    // reflexive clause: A(x, x) :- A(x, _).
    auto reflexiveClause = new AstClause();
    auto reflexiveClauseHead = new AstAtom(rel.getName());
    reflexiveClauseHead->addArgument(std::unique_ptr<AstArgument>(new AstVariable("x")));
    reflexiveClauseHead->addArgument(std::unique_ptr<AstArgument>(new AstVariable("x")));

    auto reflexiveClauseBody = new AstAtom(rel.getName());
    reflexiveClauseBody->addArgument(std::unique_ptr<AstArgument>(new AstVariable("x")));
    reflexiveClauseBody->addArgument(std::unique_ptr<AstArgument>(new AstUnnamedVariable()));

    reflexiveClause->setHead(std::unique_ptr<AstAtom>(reflexiveClauseHead));
    reflexiveClause->addToBody(std::unique_ptr<AstLiteral>(reflexiveClauseBody));
    rel.addClause(std::unique_ptr<AstClause>(reflexiveClause));
}

bool ProvenanceTransformer::transform(AstTranslationUnit& translationUnit) {
    auto program = translationUnit.getProgram();

    // get next level number
    auto getNextLevelNumber = [&](std::vector<AstArgument*> levels) {
        if (levels.size() == 0) {
            return static_cast<AstArgument*>(new AstNumberConstant(0));
        }

        if (levels.size() == 1) {
            return static_cast<AstArgument*>(
                    new AstBinaryFunctor(BinaryOp::ADD, std::unique_ptr<AstArgument>(levels[0]),
                            std::unique_ptr<AstArgument>(new AstNumberConstant(1))));
        }

        auto currentMax = new AstBinaryFunctor(BinaryOp::MAX, std::unique_ptr<AstArgument>(levels[0]),
                std::unique_ptr<AstArgument>(levels[1]));

        for (size_t i = 2; i < levels.size(); i++) {
            currentMax = new AstBinaryFunctor(BinaryOp::MAX, std::unique_ptr<AstArgument>(currentMax),
                    std::unique_ptr<AstArgument>(levels[i]));
        }

        return static_cast<AstArgument*>(
                new AstBinaryFunctor(BinaryOp::ADD, std::unique_ptr<AstArgument>(currentMax),
                        std::unique_ptr<AstArgument>(new AstNumberConstant(1))));
    };

    for (auto relation : program->getRelations()) {
        if (relation->isEqRel()) {
            transformEqrelRelation(*relation);
        }

        relation->addAttribute(std::unique_ptr<AstAttribute>(
                new AstAttribute(std::string("@rule_number"), AstTypeIdentifier("number"))));
        relation->addAttribute(std::unique_ptr<AstAttribute>(
                new AstAttribute(std::string("@level_number"), AstTypeIdentifier("number"))));

        // record clause number
        size_t clauseNum = 1;
        for (auto clause : relation->getClauses()) {
            clause->setClauseNum(clauseNum);

            // mapper to add two provenance columns to atoms
            struct M : public AstNodeMapper {
                using AstNodeMapper::operator();

                std::unique_ptr<AstNode> operator()(std::unique_ptr<AstNode> node) const override {
                    // add provenance columns
                    if (auto atom = dynamic_cast<AstAtom*>(node.get())) {
                        atom->addArgument(std::unique_ptr<AstArgument>(new AstUnnamedVariable()));
                        atom->addArgument(std::unique_ptr<AstArgument>(new AstUnnamedVariable()));
                    } else if (auto neg = dynamic_cast<AstNegation*>(node.get())) {
                        auto atom = neg->getAtom();
                        atom->addArgument(std::unique_ptr<AstArgument>(new AstUnnamedVariable()));
                        atom->addArgument(std::unique_ptr<AstArgument>(new AstUnnamedVariable()));
                    }

                    // otherwise - apply mapper recursively
                    node->apply(*this);
                    return node;
                }
            };

            // add unnamed vars to each atom nested in arguments of head
            clause->getHead()->apply(M());

            // if fact, level number is 0
            if (clause->isFact()) {
                clause->getHead()->addArgument(std::unique_ptr<AstArgument>(new AstNumberConstant(0)));
                clause->getHead()->addArgument(std::unique_ptr<AstArgument>(new AstNumberConstant(0)));
            } else {
                std::vector<AstArgument*> bodyLevels;

                for (size_t i = 0; i < clause->getBodyLiterals().size(); i++) {
                    auto lit = clause->getBodyLiterals()[i];

                    // add unnamed vars to each atom nested in arguments of lit
                    lit->apply(M());

                    // add two provenance columns to lit; first is rule num, second is level num
                    if (auto atom = dynamic_cast<AstAtom*>(lit)) {
                        atom->addArgument(std::unique_ptr<AstArgument>(new AstUnnamedVariable()));
                        atom->addArgument(std::unique_ptr<AstArgument>(
                                new AstVariable("@level_num_" + std::to_string(i))));
                        bodyLevels.push_back(new AstVariable("@level_num_" + std::to_string(i)));
                        /*
                    } else if (auto neg = dynamic_cast<AstNegation*>(lit)) {
                        auto atom = neg->getAtom();
                        atom->addArgument(std::unique_ptr<AstArgument>(new AstUnnamedVariable()));
                        atom->addArgument(std::unique_ptr<AstArgument>(new AstUnnamedVariable()));
                        */
                    }
                }

                // add two provenance columns to head lit
                clause->getHead()->addArgument(
                        std::unique_ptr<AstArgument>(new AstNumberConstant(clauseNum)));
                clause->getHead()->addArgument(std::unique_ptr<AstArgument>(getNextLevelNumber(bodyLevels)));

                clauseNum++;

                // add info relation
                program->addRelation(makeInfoRelation(*clause, translationUnit));
            }
        }
    }

    // program->print(std::cout);
    // std::cout << std::endl;

    return true;
}

}  // end of namespace souffle
